"""
Simple file which stubs out type hints for a module.

Various assumptions are baked in, which make this unsuitable for non
reticulator projects. 
"""

    # @property
    # def entities(self) -> list[EntityFileBP]: ...
    # def get_component(self, id: str) -> EntityComponentBP: ...
    # def add_component(self, name: str, data: dict) -> EntityComponentBP: ...

import sys
import importlib
import inspect
ret = importlib.import_module("reticulator")

LINES = [
	"from __future__ import annotations",
	"from core import *",
	"from typing import Tuple",
	""
] 
LINES = [x + "\n" for x in LINES]

def getmember(cls_name):
	for member in inspect.getmembers(ret):
		if member[0] == cls_name:
			return member[1]



class DecoratorManager():
	def __init__(self, name):
		self.name = name
		self.cache = []
		self.is_tracking = False

	def process_line(self, line):
		if line.startswith(self.name):
			self.is_tracking = True

		elif self.is_tracking:
			if ")" in line:
				self.is_tracking = False
			else:
				self.cache.append(line.strip().strip(",")+ "\n")

	def render(self):
		out = self.render_internal(self.cache)
		self.cache = []
		return out
	
	def render_internal(elements):
		pass
	

class ImplementResourceManager(DecoratorManager):
	def generate_decorator(self, cls_name):
		cls_name = cls_name.strip()
		cls = getmember(cls_name.strip())
		type_info = cls.type_info

		out = [
			f"@property\n",
			f"def {type_info.plural}(self) -> list[{cls_name}]:",
			f"def get_{type_info.attribute}(self, id: str) -> {cls_name}:",
			f"def add_{type_info.attribute}(self, name: str, data: dict) -> {cls_name}:"
		]

		# Handle special case for child classes
		child_cls = type_info.child_cls
		if child_cls != None:
			child_type_info = child_cls.type_info
			child_cls_name = child_cls.__name__

			out.extend([
				f"@property\n",
				f"def {child_type_info.plural}(self) -> list[{child_cls_name}]:",
				f"def get_{child_type_info.attribute}(self, id: str) -> {child_cls_name}:"
			])
			
		return ["    " + x for x in out]

	def render_internal(self, elements):
		out = []
		for element in elements:
			out.extend(self.generate_decorator(element))
		
		return out
	
class ImplementSubResourceManager(DecoratorManager):
	def generate_decorator(self, cls_name):

		cls_name = cls_name.strip()

		cls = getmember(cls_name)
		type_info = cls.type_info

		out = [
			f"@property\n",
			f"def {type_info.plural}(self) -> list[{cls_name}]:",
			f"def get_{type_info.attribute}(self, id: str) -> {cls_name}:",
			f"def add_{type_info.attribute}(self, name: str, data: dict) -> {cls_name}:"
		]
			
		return ["    " + x for x in out]

	def render_internal(self, elements):
		out = []
		for element in elements:
			out.extend(self.generate_decorator(element))
		
		return out
	
class ImplementSingleResourceManager(DecoratorManager):
	def generate_decorator(self, cls_name):

		cls_name = cls_name.strip()
		cls = getmember(cls_name)
		type_info = cls.type_info

		out = [
			f"@property\n",
			f"def {type_info.attribute}(self) -> list[{cls_name}]:",
		]
			
		return ["    " + x for x in out]

	def render_internal(self, elements):
		out = []
		for element in elements:
			out.extend(self.generate_decorator(element))
		
		return out

def main1():
	try:
		file_name = sys.argv[1]
	except IndexError:
		file_name = "./reticulator.py"

	reading_block_comment = False

	def should_take_line(line: str) -> bool:
		nonlocal reading_block_comment
		if '"""' in line:
			reading_block_comment = not reading_block_comment
			return True
		
		if reading_block_comment:
			return True
		
		strip_line = line.lstrip()
		if strip_line.startswith("class") or strip_line.startswith("def"):
			return True

	def get_indentation(s: str) -> int:
		return len(s)-len(s.lstrip(' '))


	def wants_dots(a: str, b: str) -> bool:
		# nonlocal decorator_cache

		# if decorator_cache:
		# 	return False

		strip_a = a.lstrip()
		if (strip_a.startswith("class") or strip_a.startswith("def")) and not strip_a.strip().endswith('pass'):
			# Check indentation difference
			return get_indentation(a) >= get_indentation(b)
		else:
			return False

	
	resource_manager = ImplementResourceManager("@ImplementResource(")
	sub_resource_manager = ImplementSubResourceManager("@ImplementSubResource(")
	single_manager = ImplementSingleResourceManager("@ImplementSingleResource(")

	with open(file_name, 'r') as f:
		line = "True" # Hmm
		while line:
			line = f.readline()

			resource_manager.process_line(line)
			sub_resource_manager.process_line(line)
			single_manager.process_line(line)

			if should_take_line(line):
				LINES.append(line)

				if line.lstrip().startswith("class"):
					LINES.extend(resource_manager.render())
					LINES.extend(sub_resource_manager.render())
					LINES.extend(single_manager.render())


	# Add .... where required
	for i in range(len(LINES)):
		try:
			a = LINES[i]
			b = LINES[i + 1]

			if wants_dots(a, b):
				LINES[i] = a.rstrip() + " ...\n"
			
		except Exception:
			continue

	with open('reticulator.pyi', 'w') as f:
		f.writelines(LINES)
	
main1()


from __future__ import annotations
from core import *
from typing import Tuple

def ImplementFormatVersion(jsonpath: str = "format_version"):
    """
    Class Decorator which inserts a 'format_version' property, with proper
    semantics.
    """
    def inner_format_version(cls):
        def format_version(self) -> FormatVersion: ...
        def format_version(self, format_version): ...
def ImplementIdentifier(jsonpath: str):
    """
    Class Decorator, which injects handling for accessing 'identifier'.
    :param jsonpath: The jsonpath where the identifier can be located.
    """
    def inner(cls):
        def identifier(self) -> str: ...
        def identifier(self, identifier): ...
class Pack():
    """
    Pack is a parent class that contains the shared functionality between the
    resource pack and the behavior pack.
    """
    def __init__(self, input_directory: str, project : Project = None): ...
    def project(self) -> Project:
        """
        Get the project that this pack is part of. May be None.
        """
    def save(self, force=False):
        """
        Saves every child resource.
        """
    def register_resource(self, resource: Resource) -> None:
        """
        Register a child resource to this pack. These resources will be saved
        during save.
        """
    def get_language_file(self, filepath:str) -> LanguageFile:
        """
        Gets a specific language file, based on the name of the language file.
        For example, 'texts/en_GB.lang'
        """
    def language_files(self) -> list[LanguageFile]:
        """
        Returns a list of LanguageFiles, as read from 'texts/*'
        """
class Project():
    """
    Project is a class which represents an entire 'addon', via references to a 
    ResourcePack and BehaviorPack, along with some helper methods.
    """
    def __init__(self, behavior_path: str, resource_path: str) -> Project: ...
    def set_output_directory(self, save_location: str) -> None:
        """
        Sets the save location of the RP and the BP, based on the folder
        name from their input path.

        In other words, pass in a folder where you want both the RP and the BP 
        to be saved.

        If you need finer control, set the `output_directory` on both the RP
        and the BP individually.
        """
    def get_packs(self) -> Tuple[behavior_pack, resource_pack]:
        """
        Returns Behavior Pack followed by ResourcePack.
        Useful for quickly defining rp and bp:
        rp, bp = Project("...").get_packs()
        """
    def resource_pack(self) -> ResourcePack:
        """
        The resource pack of the project.
        """
    def behavior_pack(self) -> BehaviorPack:
        """
        The behavior pack of the project.
        """
    def save(self, force=False):
        """
        Saves both packs.
        """
class LanguageFile(FileResource):
    """
    A LanguageFile is a language file, such as 'en_US.lang', and is made
    up of many Translations.
    """
    def __init__(self, filepath: str = None, pack: Pack = None) -> None: ...
    def get_translation(self, key: str) -> Translation:
        """
        Whether the language file contains the specified key.
        """
    def contains_translation(self, key: str) -> bool:
        """
        Whether the language file contains the specified key.
        """
    def delete_translation(self, key: str) -> None:
        """
        Deletes a translation based on key, if it exists.
        """
    def add_translation(self, translation: Translation, overwrite: bool = True) -> bool:
        """
        Adds a new translation key. Overwrites by default.
        """
    def _save(self): ...
    def translations(self) -> list[Translation]: ...
class Translation():
    """
    Dataclass for a translation. Many translations together make up a
    TranslationFile.
    """
    def __init__(self, key: str, value: str, comment: str = "") -> None: ...
class FormatVersion():
    def __init__(self, version) -> None: ...
    def __repr__(self) -> str: ...
    def __eq__(self, other): ...
    def __gt__(self, other): ...
class AnimationControllerBP(JsonSubResource): ...
class AnimationControllerFileBP(JsonFileResource):
    @property
    def animation_controllers(self) -> list[AnimationControllerBP]: ...
    def get_animation_controller(self, id: str) -> AnimationControllerBP: ...
    def add_animation_controller(self, name: str, data: dict) -> AnimationControllerBP: ...
class FunctionFile(FileResource):
    """
    A FunctionFile is a function file, such as run.mcfunction, and is
    made up of many commands.
    """
    def __init__(self, *args, **kwargs) -> None: ...
    def strip_comments(self):
        """
        Strips all comments from the function file.
        Generally should be used before accessing and using the `commands` property.
        """
    def commands(self) -> list[Command]:
        """
        The list of commands in this function file. Every line represents a Command.
        """
    def _save(self) -> None:
        """
        Writes the commands back to the file, one command at a time.
        """
class FeatureRuleFile(JsonFileResource): ...
class FeatureFile(JsonFileResource): ...
class SpawnRuleFile(JsonFileResource): ...
class DialogueFile(JsonFileResource): ...
class RecipeFile(JsonFileResource): ...
class EntityComponentBP(JsonSubResource): ...
class EntityEventBP(JsonSubResource): ...
class Component(JsonSubResource): ...
class ComponentGroup(JsonSubResource):
    @property
    def components(self) -> list[Component]: ...
    def get_component(self, id: str) -> Component: ...
    def add_component(self, name: str, data: dict) -> Component: ...
    """
    A component group is a collection of components in an EntityFileBP.
    """
class EntityFileBP(JsonFileResource):
    @property
    def events(self) -> list[EntityEventBP]: ...
    def get_event(self, id: str) -> EntityEventBP: ...
    def add_event(self, name: str, data: dict) -> EntityEventBP: ...
    @property
    def components(self) -> list[EntityComponentBP]: ...
    def get_component(self, id: str) -> EntityComponentBP: ...
    def add_component(self, name: str, data: dict) -> EntityComponentBP: ...
    @property
    def component_groups(self) -> list[ComponentGroup]: ...
    def get_component_group(self, id: str) -> ComponentGroup: ...
    def add_component_group(self, name: str, data: dict) -> ComponentGroup: ...
    def counterpart(self) -> EntityFileRP: ...
class LootTablePool(JsonSubResource): ...
class LootTableFile(JsonFileResource):
    @property
    def pools(self) -> list[LootTablePool]: ...
    def get_pool(self, id: str) -> LootTablePool: ...
    def add_pool(self, name: str, data: dict) -> LootTablePool: ...
class ItemComponentBP(JsonSubResource): ...
class ItemEventBP(JsonSubResource): ...
class ItemFileBP(JsonFileResource):
    @property
    def components(self) -> list[ItemComponentBP]: ...
    def get_component(self, id: str) -> ItemComponentBP: ...
    def add_component(self, name: str, data: dict) -> ItemComponentBP: ...
    @property
    def events(self) -> list[ItemEventBP]: ...
    def get_event(self, id: str) -> ItemEventBP: ...
    def add_event(self, name: str, data: dict) -> ItemEventBP: ...
class BlockFileComponentBP(JsonSubResource): ...
class BlockFileBP(JsonFileResource): ...
class AnimationBP(JsonSubResource): ...
class AnimationFileBP(JsonFileResource): ...
class BehaviorPack(Pack):
    @property
    def functions(self) -> list[FunctionFile]: ...
    def get_function(self, id: str) -> FunctionFile: ...
    def add_function(self, name: str, data: dict) -> FunctionFile: ...
    @property
    def feature_rules(self) -> list[FeatureRuleFile]: ...
    def get_feature_rule(self, id: str) -> FeatureRuleFile: ...
    def add_feature_rule(self, name: str, data: dict) -> FeatureRuleFile: ...
    @property
    def features(self) -> list[FeatureFile]: ...
    def get_feature(self, id: str) -> FeatureFile: ...
    def add_feature(self, name: str, data: dict) -> FeatureFile: ...
    @property
    def spawn_rules(self) -> list[SpawnRuleFile]: ...
    def get_spawn_rule(self, id: str) -> SpawnRuleFile: ...
    def add_spawn_rule(self, name: str, data: dict) -> SpawnRuleFile: ...
    @property
    def recipes(self) -> list[RecipeFile]: ...
    def get_recipe(self, id: str) -> RecipeFile: ...
    def add_recipe(self, name: str, data: dict) -> RecipeFile: ...
    @property
    def entities(self) -> list[EntityFileBP]: ...
    def get_entity(self, id: str) -> EntityFileBP: ...
    def add_entity(self, name: str, data: dict) -> EntityFileBP: ...
    @property
    def loot_tables(self) -> list[LootTableFile]: ...
    def get_loot_table(self, id: str) -> LootTableFile: ...
    def add_loot_table(self, name: str, data: dict) -> LootTableFile: ...
    @property
    def items(self) -> list[ItemFileBP]: ...
    def get_item(self, id: str) -> ItemFileBP: ...
    def add_item(self, name: str, data: dict) -> ItemFileBP: ...
    @property
    def blocks(self) -> list[BlockFileBP]: ...
    def get_block(self, id: str) -> BlockFileBP: ...
    def add_block(self, name: str, data: dict) -> BlockFileBP: ...
    @property
    def animation_controller_files(self) -> list[AnimationControllerFileBP]: ...
    def get_animation_controller_file(self, id: str) -> AnimationControllerFileBP: ...
    def add_animation_controller_file(self, name: str, data: dict) -> AnimationControllerFileBP: ...
    @property
    def animation_controllers(self) -> list[AnimationControllerBP]: ...
    def get_animation_controller(self, id: str) -> AnimationControllerBP: ...
    @property
    def animation_files(self) -> list[AnimationFileBP]: ...
    def get_animation_file(self, id: str) -> AnimationFileBP: ...
    def add_animation_file(self, name: str, data: dict) -> AnimationFileBP: ...
    @property
    def animations(self) -> list[AnimationBP]: ...
    def get_animation(self, id: str) -> AnimationBP: ...
    @property
    def dialogues(self) -> list[DialogueFile]: ...
    def get_dialogue(self, id: str) -> DialogueFile: ...
    def add_dialogue(self, name: str, data: dict) -> DialogueFile: ...
    """
    The BehaviorPack represents the behavior pack of a project.
    """
class Event(JsonSubResource):
    def __init__(self, data: dict = None, parent: Resource = None, json_path: str = None ) -> None: ...
    def groups_to_add(self) -> list[ComponentGroup]: ...
    def groups_to_remove(self) -> list[ComponentGroup]: ...
class Command(Resource):
    """
    A command is a wrapper around a string, which represents a single command
    in a function file.

    To use this class, you can access the 'data' property, and treat it like
    a string.
    """
    def __init__(self, command: str, file: FileResource = None, pack: Pack = None) -> None: ...
    def data(self): ...
    def data(self, data): ...
    def dirty(self): ...
    def dirty(self, dirty):
        """
        When a command is marked as dirty, it must propagate this to the
        file, so that the file can be marked as dirty.
        """
    def is_comment(self): ...
    def __str__(self): ...
    def __repr__(self): ...
class ParticleFileComponent(JsonSubResource): ...
class ParticleFileEvent(JsonSubResource): ...
class ParticleFile(JsonFileResource):
    @property
    def components(self) -> list[ParticleFileComponent]: ...
    def get_component(self, id: str) -> ParticleFileComponent: ...
    def add_component(self, name: str, data: dict) -> ParticleFileComponent: ...
    @property
    def events(self) -> list[ParticleFileEvent]: ...
    def get_event(self, id: str) -> ParticleFileEvent: ...
    def add_event(self, name: str, data: dict) -> ParticleFileEvent: ...
    """
    ParticleFile is a JsonFileResource which represents a particle file.
    """
class AnimationControllerStateRP(JsonSubResource): ...
class AnimationControllerRP(JsonSubResource):
    @property
    def states(self) -> list[AnimationControllerStateRP]: ...
    def get_state(self, id: str) -> AnimationControllerStateRP: ...
    def add_state(self, name: str, data: dict) -> AnimationControllerStateRP: ...
class AnimationControllerFileRP(JsonFileResource):
    @property
    def animation_controllers(self) -> list[AnimationControllerRP]: ...
    def get_animation_controller(self, id: str) -> AnimationControllerRP: ...
    def add_animation_controller(self, name: str, data: dict) -> AnimationControllerRP: ...
    """
    AnimationControllerFileRP
    """
class AnimationRP(JsonSubResource): ...
class AnimationFileRP(JsonFileResource):
    @property
    def animations(self) -> list[AnimationRP]: ...
    def get_animation(self, id: str) -> AnimationRP: ...
    def add_animation(self, name: str, data: dict) -> AnimationRP: ...
    """
    AnimationFileRP is a class which represents a resource pack's animation file.
    Since many animations are defined in the same file, it is often more useful.
    to use the AnimationRP class instead.
    """
class AttachableFileRP(JsonFileResource): ...
class BiomesClientFile(JsonFileResource):
    """
    BiomesClientFile is a class which represents the data stored in 'rp/biomes_client.json'
    """
class BlocksFile(JsonFileResource):
    """
    BlocksFile is a class which represents the data stored in 'rp/blocks.json'
    """
class EntityFileRP(JsonFileResource):
    @property
    def events(self) -> list[EntityEventBP]: ...
    def get_event(self, id: str) -> EntityEventBP: ...
    def add_event(self, name: str, data: dict) -> EntityEventBP: ...
    """
    EntityFileRP is a class which represents a resource pack's entity file.
    """
    def counterpart(self) -> EntityFileBP: ...
    def animations(self) -> list[AnimationTriple]: ...
    def get_animation(self, identifier:str) -> AnimationTriple:
        """
        Fetches an AnimationTriple resource, either by shortname, or identifier.
        """
    def textures(self) -> list[TextureDouble]: ...
    def get_texture(self, identifier:str) -> TextureDouble:
        """
        Fetches a texture resource, either by shortname, or texture_path.
        """
    def models(self) -> list[ModelTriple]: ...
    def get_model(self, identifier:str) -> ModelTriple:
        """
        Fetches a model resource, either by shortname, or identifier.
        """
    def materials(self) -> list[MaterialTriple]: ...
    def get_material(self, identifier:str) -> MaterialTriple:
        """
        Fetches a material resource, either by shortname, or material type.
        """
class FlipbookTexturesFile(JsonFileResource):
    """
    FlipbookTexturesFile is a class which represents the data stored in 'rp/textures/flipbook_textures.json'
    """
class FogDistanceComponent(JsonSubResource): ...
class FogVolumetricDensityComponent(JsonSubResource): ...
class FogVolumetricMediaCoefficient(JsonSubResource): ...
class FogFile(JsonFileResource):
    @property
    def distance_components(self) -> list[FogDistanceComponent]: ...
    def get_distance_component(self, id: str) -> FogDistanceComponent: ...
    def add_distance_component(self, name: str, data: dict) -> FogDistanceComponent: ...
    @property
    def volumetric_density_components(self) -> list[FogVolumetricDensityComponent]: ...
    def get_volumetric_density_component(self, id: str) -> FogVolumetricDensityComponent: ...
    def add_volumetric_density_component(self, name: str, data: dict) -> FogVolumetricDensityComponent: ...
    @property
    def volumetric_media_coefficients(self) -> list[FogVolumetricMediaCoefficient]: ...
    def get_volumetric_media_coefficient(self, id: str) -> FogVolumetricMediaCoefficient: ...
    def add_volumetric_media_coefficient(self, name: str, data: dict) -> FogVolumetricMediaCoefficient: ...
class ItemComponentRP(JsonSubResource): ...
class ItemFileRP(JsonFileResource):
    @property
    def componentss(self) -> list[ItemComponentRP]: ...
    def get_components(self, id: str) -> ItemComponentRP: ...
    def add_components(self, name: str, data: dict) -> ItemComponentRP: ...
class Material(JsonSubResource):
    """
    Represents a single material, from a .material file
    """
class MaterialFile(JsonFileResource):
    @property
    def materials(self) -> list[Material]: ...
    def get_material(self, id: str) -> Material: ...
    def add_material(self, name: str, data: dict) -> Material: ...
    """
    MaterialFile is a class which represents a resource pack's material file.
    Since many materials can be defined in the same file, it is often more useful
    to use the MaterialRP class directly.
    """
class Cube(JsonSubResource): ...
class Bone(JsonSubResource):
    @property
    def cubess(self) -> list[Cube]: ...
    def get_cubes(self, id: str) -> Cube: ...
    def add_cubes(self, name: str, data: dict) -> Cube: ...
class Model(JsonSubResource):
    @property
    def bones(self) -> list[Bone]: ...
    def get_bone(self, id: str) -> Bone: ...
    def add_bone(self, name: str, data: dict) -> Bone: ...
class ModelFile(JsonFileResource):
    @property
    def models(self) -> list[Model]: ...
    def get_model(self, id: str) -> Model: ...
    def add_model(self, name: str, data: dict) -> Model: ...
class RenderController(JsonSubResource):
    """
    A JsonSubResource, representing a single Render Controller object, contained
    within a RenderControllerFile.
    """
class RenderControllerFile(JsonFileResource):
    @property
    def render_controllers(self) -> list[RenderController]: ...
    def get_render_controller(self, id: str) -> RenderController: ...
    def add_render_controller(self, name: str, data: dict) -> RenderController: ...
class SoundDefinitionsFile(JsonFileResource):
    """
    SoundsDefinitionFile is a class which represents the data stored in
    'rp/sounds/sound_definitions.json'
    """
class SoundsFile(JsonFileResource):
    """
    SoundsFile is a class which represents the data stored in 'rp/sounds.json'
    """
class StandAloneTextureFile(JsonFileResource):
    """
    StandAloneTextureFile is a class which represents the data stored in 'rp/textures/*_texture.json'.
    Examples: 'item_texture.json', 'terrain_texture.json.

    These actual children are subclassed
    """
    def __init__(self, data: dict = None, filepath: str = None, pack: Pack = None) -> None: ...
    def texture_definitions(self) -> list[TextureFileDouble]: ...
    def get_texture_definition(self, shortname: str) -> TextureFileDouble: ...
    def add_texture_definition(self, shortname: str, textures: list[str]): ...
class TerrainTextureFile(StandAloneTextureFile): ...
class ItemTextureFile(StandAloneTextureFile): ...
class ResourceTriple(JsonSubResource):
    """
    Base class for handling "shortname": "identifier" pairs, with an underlying, resource.
    """
    def shortname(self):
        """
        This represents the shortname of the resource. e.g., "shortname": "identifier"
        """
    def shortname(self, shortname): ...
    def identifier(self):
        """
        This represents the identifier of the resource. e.g., "shortname": "identifier"
        """
    def identifier(self, identifier): ...
    def resource():
        """
        Returns the identifier associated with the resource.
        """
class MaterialTriple(ResourceTriple):
    """
    A special sub-resource, which represents a material within an RP entity.
    """
    def resource(self): ...
class AnimationTriple(ResourceTriple):
    """
    A special sub-resource, which represents an animation within an RP entity.
    """
    def resource(self): ...
class ModelTriple(ResourceTriple):
    """
    A special sub-resource, which represents a model within an RP entity.
    """
    def resource(self): ...
class TextureDouble(JsonSubResource):
    """
    A special sub-resource, which represents a texture within an RP entity.
    """
    def __init__(self, data: dict = None, parent: Resource = None, json_path: str = None ) -> None: ...
    def shortname(self): ...
    def shortname(self, shortname): ...
    def texture_path(self): ...
    def texture_path(self, texture_path): ...
    def exists(self) -> bool:
        """
        Returns True if this resource exists in the pack.
        """
class TextureFileDouble(JsonSubResource):
    """
    A special sub-resource, which represents a texture within a texture
    definition file, such as 'item_texture.json' ...

    This is a special case, as it has some additional logic for obscuring the
    texture path. This is because textures are stored nested under 
    'textures' which is simply inconvenient to work with.
    """
    def __init__(self, data: dict = None, parent: Resource = None, json_path: str = None ) -> None: ...
    def shortname(self): ...
    def shortname(self, shortname): ...
    def data(self):
        """
        Custom data getter allows us to re-create the json structure based
        on the saved textures.
        """
class ResourcePack(Pack):
    @property
    def particles(self) -> list[ParticleFile]: ...
    def get_particle(self, id: str) -> ParticleFile: ...
    def add_particle(self, name: str, data: dict) -> ParticleFile: ...
    @property
    def attachables(self) -> list[AttachableFileRP]: ...
    def get_attachable(self, id: str) -> AttachableFileRP: ...
    def add_attachable(self, name: str, data: dict) -> AttachableFileRP: ...
    @property
    def entities(self) -> list[EntityFileRP]: ...
    def get_entity(self, id: str) -> EntityFileRP: ...
    def add_entity(self, name: str, data: dict) -> EntityFileRP: ...
    @property
    def fogs(self) -> list[FogFile]: ...
    def get_fog(self, id: str) -> FogFile: ...
    def add_fog(self, name: str, data: dict) -> FogFile: ...
    @property
    def itemss(self) -> list[ItemFileRP]: ...
    def get_items(self, id: str) -> ItemFileRP: ...
    def add_items(self, name: str, data: dict) -> ItemFileRP: ...
    @property
    def animation_controller_files(self) -> list[AnimationControllerFileRP]: ...
    def get_animation_controller_file(self, id: str) -> AnimationControllerFileRP: ...
    def add_animation_controller_file(self, name: str, data: dict) -> AnimationControllerFileRP: ...
    @property
    def animation_controllers(self) -> list[AnimationControllerRP]: ...
    def get_animation_controller(self, id: str) -> AnimationControllerRP: ...
    @property
    def animation_files(self) -> list[AnimationFileRP]: ...
    def get_animation_file(self, id: str) -> AnimationFileRP: ...
    def add_animation_file(self, name: str, data: dict) -> AnimationFileRP: ...
    @property
    def animations(self) -> list[AnimationRP]: ...
    def get_animation(self, id: str) -> AnimationRP: ...
    @property
    def material_files(self) -> list[MaterialFile]: ...
    def get_material_file(self, id: str) -> MaterialFile: ...
    def add_material_file(self, name: str, data: dict) -> MaterialFile: ...
    @property
    def materials(self) -> list[Material]: ...
    def get_material(self, id: str) -> Material: ...
    @property
    def model_files(self) -> list[ModelFile]: ...
    def get_model_file(self, id: str) -> ModelFile: ...
    def add_model_file(self, name: str, data: dict) -> ModelFile: ...
    @property
    def models(self) -> list[Model]: ...
    def get_model(self, id: str) -> Model: ...
    @property
    def render_controller_files(self) -> list[RenderControllerFile]: ...
    def get_render_controller_file(self, id: str) -> RenderControllerFile: ...
    def add_render_controller_file(self, name: str, data: dict) -> RenderControllerFile: ...
    @property
    def render_controllers(self) -> list[RenderController]: ...
    def get_render_controller(self, id: str) -> RenderController: ...
    @property
    def sounds_file(self) -> list[SoundsFile]: ...
    @property
    def sound_definitions_file(self) -> list[SoundDefinitionsFile]: ...
    @property
    def flipbook_textures_file(self) -> list[FlipbookTexturesFile]: ...
    @property
    def blocks_file(self) -> list[BlocksFile]: ...
    @property
    def biomes_client_file(self) -> list[BiomesClientFile]: ...
    @property
    def terrain_texture_file(self) -> list[TerrainTextureFile]: ...
    @property
    def item_texture_file(self) -> list[ItemTextureFile]: ...
    def __init__(self, input_path: str, project: Project = None): ...
    def sounds(self) -> list[str]:
        """
        Returns a list of all sounds in the pack, relative to the pack root.
        """
    def get_sounds(self, search_path: str = "", trim_extension: bool = True) -> list[str]:
        """
        Returns a list of all child sounds of the searchpath, relative to the pack root. 
        Search path should not include 'sounds'.

        You may optionally trim the extension from the returned paths.

        Example: rp.get_sounds("entities", trim_extension=True)
        """
    def textures(self) -> list[str]:
        """
        Returns a list of all textures in the pack, relative to the pack root.

        Example: "textures/my_texture.png"
        """
    def get_textures(self, search_path: str = "", trim_extension: bool = True) -> list[str]:
        """
        Returns a list of all child textures of the searchpath, relative to the pack root. 
        Search path should not include 'textures'.

        You may optionally trim the extension from the returned paths.

        Example: rp.get_textures("entities", trim_extension=True)
        """



from __future__ import annotations

import re
import os
import glob

from functools import cached_property
from typing import Tuple

from .core import *

def ImplementFormatVersion(jsonpath: str = "format_version"):
    """
    Class Decorator which inserts a 'format_version' property, with proper
    semantics.
    """

    def inner_format_version(cls):
        @property
        def format_version(self) -> FormatVersion:
            return FormatVersion(self.get_jsonpath(jsonpath))
        
        @format_version.setter
        def format_version(self, format_version):
            return self.set_jsonpath(jsonpath, str(FormatVersion(format_version)))

        cls.format_version = format_version
        return cls
    
    return inner_format_version


def ImplementIdentifier(jsonpath: str):
    """
    Class Decorator, which injects handling for accessing 'identifier'.
    :param jsonpath: The jsonpath where the identifier can be located.
    """

    def inner(cls):
        @property
        def identifier(self) -> str:
            return self.get_jsonpath(jsonpath)
        
        @identifier.setter
        def identifier(self, identifier):
            return self.set_jsonpath(jsonpath, identifier)

        cls.identifier = identifier
        
        return cls

    return inner


class Pack():
    """
    Pack is a parent class that contains the shared functionality between the
    resource pack and the behavior pack.
    """
    def __init__(self, input_directory: str, project : Project = None):
        self.resources = []
        self._language_files = []
        self._project = project

        # The input path is the path to the folder containing the pack.
        self.input_path: str = input_directory

        # The output path is the path to the output directory, where this
        # pack will be saved. This is the same as the input path, unless
        # explicitely saved.
        self.output_directory: str = input_directory

    @cached_property
    def project(self) -> Project:
        """
        Get the project that this pack is part of. May be None.
        """
        return self._project

    def save(self, force=False):
        """
        Saves every child resource.
        """
        for resource in self.resources:
            resource.save(force=force)

    def register_resource(self, resource: Resource) -> None:
        """
        Register a child resource to this pack. These resources will be saved
        during save.
        """

        # Attempt to set the pack for this resource.
        try:
            resource.pack = self
        except Exception:
            pass
        self.resources.append(resource)

    def get_language_file(self, filepath:str) -> LanguageFile:
        """
        Gets a specific language file, based on the name of the language file.
        For example, 'texts/en_GB.lang'
        """
        for language_file in self.language_files:
            if smart_compare(language_file.filepath, filepath):
                return language_file
        raise AssetNotFoundError(filepath)

    @cached_property
    def language_files(self) -> list[LanguageFile]:
        """
        Returns a list of LanguageFiles, as read from 'texts/*'
        """
        base_directory = os.path.join(self.input_path, "texts")
        for local_path in glob.glob(base_directory + "/**/*.lang", recursive=True):
            local_path = os.path.relpath(local_path, self.input_path)
            self._language_files.append(LanguageFile(filepath = local_path, pack = self))
            
        return self._language_files

class Project():
    """
    Project is a class which represents an entire 'addon', via references to a 
    ResourcePack and BehaviorPack, along with some helper methods.
    """
    def __init__(self, behavior_path: str, resource_path: str) -> Project:
        self.__behavior_path = behavior_path
        self.__resource_path = resource_path
        self.__resource_pack : ResourcePack = None
        self.__behavior_pack : BehaviorPack = None
    
    def set_output_directory(self, save_location: str) -> None:
        """
        Sets the save location of the RP and the BP, based on the folder
        name from their input path.

        In other words, pass in a folder where you want both the RP and the BP 
        to be saved.

        If you need finer control, set the `output_directory` on both the RP
        and the BP individually.
        """
        self.resource_pack.output_directory = save_location + "/" + os.path.dirname(self.resource_pack.input_path)
        self.behavior_pack.output_directory = save_location + "/" + os.path.dirname(self.behavior_pack.input_path)

    def get_packs(self) -> Tuple[behavior_pack, resource_pack]:
        """
        Returns Behavior Pack followed by ResourcePack.
        Useful for quickly defining rp and bp:
        rp, bp = Project("...").get_packs()
        """
        return self.behavior_pack, self.resource_pack
        
    @cached_property
    def resource_pack(self) -> ResourcePack:
        """
        The resource pack of the project.
        """
        self.__resource_pack = ResourcePack(self.__resource_path, project=self)
        return self.__resource_pack

    @cached_property
    def behavior_pack(self) -> BehaviorPack:
        """
        The behavior pack of the project.
        """
        self.__behavior_pack = BehaviorPack(self.__behavior_path, project=self)
        return self.__behavior_pack

    def save(self, force=False):
        """
        Saves both packs.
        """
        self.__behavior_pack.save(force=force)
        self.__resource_pack.save(force=force)


class LanguageFile(FileResource):
    """
    A LanguageFile is a language file, such as 'en_US.lang', and is made
    up of many Translations.
    """
    def __init__(self, filepath: str = None, pack: Pack = None) -> None:
        super().__init__(filepath=filepath, pack=pack)
        self.__translations: list[Translation] = []
    
    def get_translation(self, key: str) -> Translation:
        """
        Whether the language file contains the specified key.
        """
        for translation in self.translations:
            if translation.key == key:
                return translation 
        raise AssetNotFoundError(f"Translation with key '{key}' not found in language file '{self.filepath}'.")

    def contains_translation(self, key: str) -> bool:
        """
        Whether the language file contains the specified key.
        """

        for translation in self.translations:
            if translation.key == key:
                return True
        return False

    def delete_translation(self, key: str) -> None:
        """
        Deletes a translation based on key, if it exists.
        """
        for translation in self.translations:
            if translation.key == key:
                self.dirty = True
                self.translations.remove(translation)
                return

    def add_translation(self, translation: Translation, overwrite: bool = True) -> bool:
        """
        Adds a new translation key. Overwrites by default.
        """

        # We must complain about duplicates.
        # If no overwrite, don't add
        # If overwrite, delete previous translation
        if self.contains_translation(translation.key):
            if not overwrite:
                return False
            else:
                self.delete_translation(translation.key)

        self.dirty = True
        self.__translations.append(translation)

        return True

    def _save(self):
        path = os.path.join(self.pack.output_directory, self.filepath)
        create_nested_directory(path)
        with open(path, 'w', encoding='utf-8') as file:
            for translation in self.translations:
                file.write(f"{translation.key}={translation.value}\t##{translation.comment}\n")


    @cached_property
    def translations(self) -> list[Translation]:
        with open(os.path.join(self.pack.input_path, self.filepath), "r", encoding='utf-8') as language_file:
            for line in language_file.readlines():
                language_regex = "^([^#\n]+?)=([^#]+)#*?([^#]*?)$"
                if match := re.search(language_regex, line):
                    groups = match.groups()
                    self.__translations.append(
                        Translation(
                            key = groups[0].strip() if len(groups) > 0 else "",
                            value = groups[1].strip() if len(groups) > 1 else "",
                            comment = groups[2].strip() if len(groups) > 2 else "",
                        )
                    )
        return self.__translations

# Generic Resources
class Translation():
    """
    Dataclass for a translation. Many translations together make up a
    TranslationFile.
    """

    def __init__(self, key: str, value: str, comment: str = "") -> None:
        self.key = key
        self.value = value
        self.comment = comment

class FormatVersion():
    def __init__(self, version) -> None:
        if isinstance(version, FormatVersion):
            self.major = version.major
            self.minor = version.minor
            self.patch = version.patch
            return
        elif isinstance(version, str):
            elements = version.split('.')
        else:
            # Change to suitable error
            raise TypeError()

        # Pack with extra data if it's missing
        for i in range(3 - len(elements)):
            elements.append('0')
        
        self.major = int(elements[0])
        self.minor = int(elements[1])
        self.patch = int(elements[2])

    def __repr__(self) -> str:
        return f'{self.major}.{self.minor}.{self.patch}'
        
    def __eq__(self, other):
        other = FormatVersion(other)
        return self.major == other.major and self.minor == other.minor and self.patch == other.patch

    def __gt__(self, other):
        if self.major > other.major:
            return True
        elif self.major < other.major:
            return False

        if self.minor > other.minor:
            return True
        elif self.minor < other.minor:
            return False

        if self.patch > other.patch:
            return True
        elif self.patch < other.patch:
            return False
        
        return self != other


class AnimationControllerBP(JsonSubResource):
    type_info = TypeInfo(
        jsonpath = "animation_controllers",
        attribute = "animation_controller",
        getter_attribute = "id"
    )

@ImplementFormatVersion()
@ImplementSubResource(
    AnimationControllerBP
)
class AnimationControllerFileBP(JsonFileResource):
    type_info = TypeInfo(
        filepath = "animation_controllers",
        attribute = "animation_controller_file",
        getter_attribute = "filepath",
        child_cls = AnimationControllerBP
    )


class FunctionFile(FileResource):
    """
    A FunctionFile is a function file, such as run.mcfunction, and is
    made up of many commands.
    """
    
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._commands : list[Command] = []

    type_info = TypeInfo(
        filepath = "functions",
        extension = ".mcfunction",
        attribute = "function",
        getter_attribute="filepath"
    )
    
    def strip_comments(self):
        """
        Strips all comments from the function file.
        Generally should be used before accessing and using the `commands` property.
        """
        self.commands = [c for c in self.commands if not c.is_comment()]

    @cached_property
    def commands(self) -> list[Command]:
        """
        The list of commands in this function file. Every line represents a Command.
        """
        with open(os.path.join(self.pack.input_path, self.filepath), "r", encoding='utf-8') as function_file:
            for line in function_file.readlines():
                command = line.strip()
                if command:
                    self._commands.append(Command(command, file=self, pack=self.pack))
        self._commands = convert_to_notify_structure(self._commands, self)
        return self._commands
    
    def _save(self) -> None:
        """
        Writes the commands back to the file, one command at a time.
        """
        path = os.path.join(self.pack.output_directory, self.filepath)
        create_nested_directory(path)
        with open(path, 'w', encoding='utf-8') as file:
            for command in self.commands:
                file.write(command.data + '\n')

@ImplementIdentifier("minecraft:feature_rules/description/identifier")
@ImplementFormatVersion()
class FeatureRuleFile(JsonFileResource):
    type_info = TypeInfo(
        filepath = "feature_rules",
        attribute = "feature_rule",
        getter_attribute="identifier"
    )

@ImplementIdentifier("**/description/identifier")
@ImplementFormatVersion()
class FeatureFile(JsonFileResource):
    type_info = TypeInfo(
        filepath = "features",
        attribute = "feature",
        getter_attribute="identifier"
    )
    
@ImplementFormatVersion()
@ImplementIdentifier("minecraft:spawn_rules/description/identifier")
class SpawnRuleFile(JsonFileResource):
    type_info = TypeInfo(
        filepath = "spawn_rules",
        attribute = "spawn_rule",
        getter_attribute="identifier"
    )

@ImplementFormatVersion()
class DialogueFile(JsonFileResource):
    type_info = TypeInfo(
        filepath = "dialogue",
        attribute = "dialogue",
        plural = "dialogues",
        getter_attribute = "filepath"
    )

@ImplementIdentifier("**/identifier")
@ImplementFormatVersion()
class RecipeFile(JsonFileResource):
    type_info = TypeInfo(
        filepath = "recipes",
        attribute = "recipe",
        getter_attribute="identifier"
    )

class EntityComponentBP(JsonSubResource):
    type_info = TypeInfo(
        jsonpath = "minecraft:entity/components",
        attribute = "component",
        getter_attribute = "id"
    )

class EntityEventBP(JsonSubResource):
    type_info = TypeInfo(
        jsonpath = "minecraft:entity/events",
        attribute = "event",
        getter_attribute = "id"
    )

class Component(JsonSubResource):
    type_info = TypeInfo(
        jsonpath = "**",
        attribute = "component",
        getter_attribute = "id"
    )

@ImplementSubResource(
    Component
)
class ComponentGroup(JsonSubResource):
    """
    A component group is a collection of components in an EntityFileBP.
    """
    type_info = TypeInfo(
        jsonpath = "minecraft:entity/component_groups",
        attribute = "component_group",
        getter_attribute = "id"
    )

@ImplementFormatVersion()
@ImplementIdentifier("minecraft:entity/description/identifier")
@ImplementSubResource(
    EntityEventBP, 
    EntityComponentBP, 
    ComponentGroup
)
class EntityFileBP(JsonFileResource):
    type_info = TypeInfo(
        filepath = "entities",
        attribute = "entity",
        plural = "entities",
        getter_attribute = "identifier"
    )

    @property
    def counterpart(self) -> EntityFileRP:
        return self.pack.project.resource_pack.get_entity(self.identifier)

class LootTablePool(JsonSubResource):
    type_info = TypeInfo(
        jsonpath = "pools",
        attribute = "pool"
    )

@ImplementSubResource(
    LootTablePool
)
class LootTableFile(JsonFileResource):
    type_info = TypeInfo(
        filepath = "loot_tables",
        attribute = "loot_table",
        getter_attribute = "filepath"
    )

class ItemComponentBP(JsonSubResource):
    type_info = TypeInfo(
        jsonpath = "minecraft:item/components",
        attribute = "component",
        getter_attribute = "id"
    )

class ItemEventBP(JsonSubResource):
    type_info = TypeInfo(
        jsonpath = "minecraft:item/events",
        attribute = "event",
        getter_attribute = "id"
    )


@ImplementFormatVersion()
@ImplementIdentifier("minecraft:item/description/identifier")
@ImplementSubResource(
    ItemComponentBP, 
    ItemEventBP
)
class ItemFileBP(JsonFileResource):
    type_info = TypeInfo(
        filepath = "items",
        attribute = "item",
        getter_attribute = "identifier"
    )

class BlockFileComponentBP(JsonSubResource):
    type_info = TypeInfo(
        jsonpath = "minecraft:block/components",
        attribute = "component",
        getter_attribute = "id"
    )


@ImplementFormatVersion()
@ImplementIdentifier(jsonpath="minecraft:block/description/identifier")
@ImplementSubResource(BlockFileComponentBP)
class BlockFileBP(JsonFileResource):
    type_info = TypeInfo(
        filepath = "blocks",
        attribute = "block",
        getter_attribute = "identifier"
    )

@ClassProperty('loop')
class AnimationBP(JsonSubResource):
    type_info = TypeInfo(
        jsonpath = "animations",
        attribute = "animation",
        getter_attribute = "id"
    )

@ImplementSubResource(AnimationBP)
class AnimationFileBP(JsonFileResource):
    type_info = TypeInfo(
        filepath = "animations",
        attribute = "animation_file",
        getter_attribute = "filepath",
        child_cls=AnimationBP
    )

@ImplementResource(
    FunctionFile,
    FeatureRuleFile,
    FeatureFile,
    SpawnRuleFile,
    RecipeFile,
    EntityFileBP,
    LootTableFile,
    ItemFileBP,
    BlockFileBP,
    AnimationControllerFileBP,
    AnimationFileBP,
    DialogueFile
)
class BehaviorPack(Pack):
    """
    The BehaviorPack represents the behavior pack of a project.
    """

class Event(JsonSubResource):
    def __init__(self, data: dict = None, parent: Resource = None, json_path: str = None ) -> None:
        super().__init__(data=data, parent=parent, json_path=json_path)
        
        self.__groups_to_add = []
        self.__groups_to_remove = []

    @cached_property
    def groups_to_add(self) -> list[ComponentGroup]:
        for path, data in self.get_data_at("add/component_groups"):
            self.__groups_to_add.append(ComponentGroup(parent = self, json_path = path, data = data))
        return self.__groups_to_add
    
    @cached_property
    def groups_to_remove(self) -> list[ComponentGroup]:
        for path, data in self.get_data_at("remove/component_groups"):
            self.__groups_to_remove.append(ComponentGroup(parent = self, json_path = path, data = data))
        return self.__groups_to_remove

class Command(Resource):
    """
    A command is a wrapper around a string, which represents a single command
    in a function file.

    To use this class, you can access the 'data' property, and treat it like
    a string.
    """
    def __init__(self, command: str, file: FileResource = None, pack: Pack = None) -> None:
        super().__init__(file=file, pack=pack)

        # The 'data' is the actual command, which is stored as a string.
        self._data: str = command

    @property
    def data(self):
        return self._data
    
    @data.setter
    def data(self, data):
        self._data = data
        self.dirty = True

    @property
    def dirty(self):
        return self._dirty

    @dirty.setter
    def dirty(self, dirty):
        """
        When a command is marked as dirty, it must propagate this to the
        file, so that the file can be marked as dirty.
        """
        self._dirty = dirty
        self.file.dirty = dirty

    def is_comment(self):
        return self.data.startswith("#")

    def __str__(self):
        return self.data

    def __repr__(self):
        return f"Command: '{self.data}'"

class ParticleFileComponent(JsonSubResource):
    type_info = TypeInfo(
        jsonpath = "particle_effect/components",
        attribute = "component",
        getter_attribute = "id"
    )

class ParticleFileEvent(JsonSubResource):
    type_info = TypeInfo(
        jsonpath = "particle_effect/events",
        attribute = "event",
        getter_attribute = "id"
    )

@ImplementFormatVersion()
@ImplementIdentifier("particle_effect/description/identifier")
@ImplementSubResource(
    ParticleFileComponent,
    ParticleFileEvent
)
class ParticleFile(JsonFileResource):
    """
    ParticleFile is a JsonFileResource which represents a particle file.
    """
    type_info = TypeInfo(
        filepath = "particles",
        attribute = "particle",
        getter_attribute = "identifier",
    )

class AnimationControllerStateRP(JsonSubResource):
    type_info = TypeInfo(
        jsonpath = "states",
        attribute = "state",
        getter_attribute = "id"
    )

@ClassProperty('initial_state')
@ImplementSubResource(
    AnimationControllerStateRP
)
class AnimationControllerRP(JsonSubResource):
    type_info = TypeInfo(
        jsonpath = "animation_controllers",
        attribute = "animation_controller",
        getter_attribute = "id"
    )

@ImplementSubResource(
    AnimationControllerRP
)
class AnimationControllerFileRP(JsonFileResource):
    """
    AnimationControllerFileRP
    """
    type_info = TypeInfo(
        filepath = "animation_controllers",
        attribute = "animation_controller_file",
        getter_attribute = "filepath",
        child_cls=AnimationControllerRP
    )

@ClassProperty('loop')
class AnimationRP(JsonSubResource):
    type_info = TypeInfo(
        jsonpath = "animations",
        attribute = "animation",
        getter_attribute = "id"
    )

@ImplementSubResource(
    AnimationRP
)
class AnimationFileRP(JsonFileResource):
    """
    AnimationFileRP is a class which represents a resource pack's animation file.
    Since many animations are defined in the same file, it is often more useful.
    to use the AnimationRP class instead.
    """

    type_info = TypeInfo(
        filepath = "animations",
        attribute = "animation_file",
        getter_attribute = "filepath",
        child_cls=AnimationRP
    )


@ImplementFormatVersion()
@ImplementIdentifier("minecraft:attachable/description/identifier")
class AttachableFileRP(JsonFileResource):
    type_info = TypeInfo(
        filepath = "attachables",
        attribute = "attachable",
        getter_attribute = "identifier"
    )


class BiomesClientFile(JsonFileResource):
    """
    BiomesClientFile is a class which represents the data stored in 'rp/biomes_client.json'
    """

    type_info = TypeInfo(
        filepath = "biomes_client",
        attribute = "biomes_client_file",
        plural = "biomes_client_file"
    )


class BlocksFile(JsonFileResource):
    """
    BlocksFile is a class which represents the data stored in 'rp/blocks.json'
    """

    type_info = TypeInfo(
        filepath = "blocks",
        attribute = "blocks_file"
    )



@ImplementFormatVersion()
@ImplementIdentifier("minecraft:client_entity/description/identifier")
@ImplementSubResource(
    EntityEventBP
)
class EntityFileRP(JsonFileResource):
    """
    EntityFileRP is a class which represents a resource pack's entity file.
    """
    type_info = TypeInfo(
        filepath = "entity",
        attribute = "entity",
        plural = "entities",
        getter_attribute = "identifier"
    )

    @property
    def counterpart(self) -> EntityFileBP:
        return self.pack.project.behavior_pack.get_entity(self.identifier)

    # TODO: See if the decoration system can apply to these 'triples'.
    # Or we need to add 'adders' for all three (anim, texture, model)
    @cached_property
    def animations(self) -> list[AnimationTriple]:
        self._animations = []
        for path, data in self.get_data_at("minecraft:client_entity/description/animations"):
            self._animations.append(AnimationTriple(parent = self, json_path = path, data = data))
        return self._animations
    
    def get_animation(self, identifier:str) -> AnimationTriple:
        """
        Fetches an AnimationTriple resource, either by shortname, or identifier.
        """
        for child in self.animations:
            if smart_compare(child.shortname, identifier) or smart_compare(child.identifier, identifier):
                return child
        raise AssetNotFoundError(id)

    @cached_property
    def textures(self) -> list[TextureDouble]:
        self._textures = []
        for path, data in self.get_data_at("minecraft:client_entity/description/textures"):
            self._textures.append(TextureDouble(parent = self, json_path = path, data = data))
        return self._textures
    
    def get_texture(self, identifier:str) -> TextureDouble:
        """
        Fetches a texture resource, either by shortname, or texture_path.
        """
        for child in self.textures:
            if smart_compare(child.shortname, identifier) or smart_compare(child.texture_path, identifier):
                return child
        raise AssetNotFoundError(id)

    @cached_property
    def models(self) -> list[ModelTriple]:
        self._models = []
        for path, data in self.get_data_at("minecraft:client_entity/description/geometry"):
            self._models.append(ModelTriple(parent = self, json_path = path, data = data))
        return self._models
    
    def get_model(self, identifier:str) -> ModelTriple:
        """
        Fetches a model resource, either by shortname, or identifier.
        """
        for child in self.models:
            if smart_compare(child.shortname, identifier) or smart_compare(child.identifier, identifier):
                return child
        raise AssetNotFoundError(id)

    @cached_property
    def materials(self) -> list[MaterialTriple]:
        self._materials = []
        for path, data in self.get_data_at("minecraft:client_entity/description/materials"):
            self._materials.append(MaterialTriple(parent = self, json_path = path, data = data))
        return self._materials
    
    def get_material(self, identifier:str) -> MaterialTriple:
        """
        Fetches a material resource, either by shortname, or material type.
        """
        for child in self.materials:
            if smart_compare(child.shortname, identifier) or smart_compare(child.identifier, identifier):
                return child
        raise AssetNotFoundError(id)


class FlipbookTexturesFile(JsonFileResource):
    """
    FlipbookTexturesFile is a class which represents the data stored in 'rp/textures/flipbook_textures.json'
    """

    type_info = TypeInfo(
        filepath = "textures/flipbook_textures",
        attribute = "flipbook_textures_file"
    )


class FogDistanceComponent(JsonSubResource):
    type_info = TypeInfo(
        jsonpath = "minecraft:fog_settings/distance",
        attribute = "distance_component",
        getter_attribute = "id"
    )

class FogVolumetricDensityComponent(JsonSubResource):
    type_info = TypeInfo(
        jsonpath = "minecraft:fog_settings/volumetric/density",
        attribute = "volumetric_density_component",
        getter_attribute = "id"
    )

class FogVolumetricMediaCoefficient(JsonSubResource):
    type_info = TypeInfo(
        jsonpath = "minecraft:fog_settings/volumetric/media_coefficients",
        attribute = "volumetric_media_coefficient",
        getter_attribute = "id"
    )

@ImplementFormatVersion()
@ImplementIdentifier("minecraft:fog_settings/description/identifier")
@ImplementSubResource(
    FogDistanceComponent,
    FogVolumetricDensityComponent,
    FogVolumetricMediaCoefficient
)
class FogFile(JsonFileResource):
    type_info = TypeInfo(
        filepath = "fogs",
        attribute = "fog",
        getter_attribute = "identifier"
    )

class ItemComponentRP(JsonSubResource):
    type_info = TypeInfo(
        jsonpath = "minecraft:item/components",
        attribute = "components",
        getter_attribute = "id"
    )

@ImplementFormatVersion()
@ImplementIdentifier("minecraft:item/description/identifier")
@ImplementSubResource(
    ItemComponentRP
)
class ItemFileRP(JsonFileResource):
    type_info = TypeInfo(
        filepath = "items",
        attribute = "items",
        getter_attribute = "filepath"
    )

class Material(JsonSubResource):
    """
    Represents a single material, from a .material file
    """

    type_info = TypeInfo(
        jsonpath = "**",
        attribute = "material",
        getter_attribute = "id"
    )

@ImplementFormatVersion(jsonpath="materials/version")
@ImplementSubResource(
    Material
)
class MaterialFile(JsonFileResource):
    """
    MaterialFile is a class which represents a resource pack's material file.
    Since many materials can be defined in the same file, it is often more useful
    to use the MaterialRP class directly.
    """
    type_info = TypeInfo(
        filepath = "materials",
        attribute = "material_file",
        getter_attribute = "filepath",
        extension="material",
        child_cls=Material
    )

class Cube(JsonSubResource):
    type_info = TypeInfo(
        jsonpath = "cubes",
        attribute = "cubes",
        getter_attribute = "id"
    )

@ClassProperty("name")
@ImplementSubResource(
    Cube
)
class Bone(JsonSubResource):
    name: str
    type_info = TypeInfo(
        jsonpath = "bones",
        attribute = "bone",
        getter_attribute = "name"
    )


@ImplementIdentifier("description/identifier")
@ImplementSubResource(
    Bone
)
class Model(JsonSubResource):
    identifier: str
    type_info = TypeInfo(
        jsonpath = "minecraft:geometry",
        attribute = "model",
        getter_attribute = "identifier"
    )

@ImplementFormatVersion()
@ImplementSubResource(
    Model
)
class ModelFile(JsonFileResource):
    type_info = TypeInfo(
        filepath = "models",
        attribute = "model_file",
        getter_attribute = "filepath",
        child_cls=Model
    )

class RenderController(JsonSubResource):
    """
    A JsonSubResource, representing a single Render Controller object, contained
    within a RenderControllerFile.
    """

    type_info = TypeInfo(
        jsonpath = "render_controllers",
        attribute = "render_controller",
        getter_attribute = "id"
    )

@ImplementFormatVersion()
@ImplementSubResource(
    RenderController
)
class RenderControllerFile(JsonFileResource):
    type_info = TypeInfo(
        filepath = "render_controllers",
        attribute = "render_controller_file",
        getter_attribute = "filepath",
        child_cls=RenderController
    )


@ImplementFormatVersion()
class SoundDefinitionsFile(JsonFileResource):
    """
    SoundsDefinitionFile is a class which represents the data stored in
    'rp/sounds/sound_definitions.json'
    """
    type_info = TypeInfo(
        filepath = "sounds/sound_definitions",
        attribute = "sound_definitions_file"
    )


class SoundsFile(JsonFileResource):
    """
    SoundsFile is a class which represents the data stored in 'rp/sounds.json'
    """
    type_info = TypeInfo(
        filepath = "sounds",
        attribute = "sounds_file"
    )

class StandAloneTextureFile(JsonFileResource):
    """
    StandAloneTextureFile is a class which represents the data stored in 'rp/textures/*_texture.json'.
    Examples: 'item_texture.json', 'terrain_texture.json.

    These actual children are subclassed
    """
    def __init__(self, data: dict = None, filepath: str = None, pack: Pack = None) -> None:
        super().__init__(data = data, filepath = filepath, pack = pack)
        self.__texture_definitions = []
    
    @cached_property
    def texture_definitions(self) -> list[TextureFileDouble]:
        for path, data in self.get_data_at("texture_data"):
            self.__texture_definitions.append(TextureFileDouble(parent = self, json_path = path, data = data))
        return self.__texture_definitions

    def get_texture_definition(self, shortname: str) -> TextureFileDouble:
        for child in self.texture_definitions:
            if child.shortname == shortname:
                return child
        raise AssetNotFoundError(f"Texture definition for shortname '{shortname}' not found.")

    def add_texture_definition(self, shortname: str, textures: list[str]):
        self.set_jsonpath(f"texture_data/{shortname}", {
            "textures": textures
        })
        self.__texture_definitions.append(TextureFileDouble(parent = self, json_path = f"texture_data/{shortname}", data = {"textures": textures}))

class TerrainTextureFile(StandAloneTextureFile):
    type_info = TypeInfo(
        filepath = "textures/terrain_texture",
        attribute = "terrain_texture_file"
    )

class ItemTextureFile(StandAloneTextureFile):
    type_info = TypeInfo(
        filepath = "textures/item_texture",
        attribute = "item_texture_file"
    )
    

class ResourceTriple(JsonSubResource):
    """
    Base class for handling "shortname": "identifier" pairs, with an underlying, resource.
    """
    @property
    def shortname(self):
        """
        This represents the shortname of the resource. e.g., "shortname": "identifier"
        """
        return self.id

    @shortname.setter
    def shortname(self, shortname):
        self.id = shortname
    
    @property
    def identifier(self):
        """
        This represents the identifier of the resource. e.g., "shortname": "identifier"
        """
        return self.data
    
    @identifier.setter
    def identifier(self, identifier):
        self.data = identifier

    def resource():
        """
        Returns the identifier associated with the resource.
        """
        raise NotImplementedError

class MaterialTriple(ResourceTriple):
    """
    A special sub-resource, which represents a material within an RP entity.
    """

    @property
    def resource(self):
        return self.parent.pack.get_material(self.identifier)


class AnimationTriple(ResourceTriple):
    """
    A special sub-resource, which represents an animation within an RP entity.
    """

    @property
    def resource(self):
        return self.parent.pack.get_animation(self.identifier)
    
class ModelTriple(ResourceTriple):
    """
    A special sub-resource, which represents a model within an RP entity.
    """

    @property
    def resource(self):
        return self.parent.pack.get_model(self.identifier)
    

class TextureDouble(JsonSubResource):
    """
    A special sub-resource, which represents a texture within an RP entity.
    """
    def __init__(self, data: dict = None, parent: Resource = None, json_path: str = None ) -> None:
        super().__init__(data=data, parent=parent, json_path=json_path)

    @property
    def shortname(self):
        return self.id

    @shortname.setter
    def shortname(self, shortname):
        self.id = shortname
    
    @property
    def texture_path(self):
        return self.data
    
    @texture_path.setter
    def texture_path(self, texture_path):
        self.data = texture_path
    
    def exists(self) -> bool:
        """
        Returns True if this resource exists in the pack.
        """
        return os.path.exists(os.path.join(self.pack.input_path, self.filepath))


class TextureFileDouble(JsonSubResource):
    """
    A special sub-resource, which represents a texture within a texture
    definition file, such as 'item_texture.json'

    This is a special case, as it has some additional logic for obscuring the
    texture path. This is because textures are stored nested under 
    'textures' which is simply inconvenient to work with.
    """

    def __init__(self, data: dict = None, parent: Resource = None, json_path: str = None ) -> None:
        super().__init__(data=data, parent=parent, json_path=json_path)
 
        self.textures = data.get("textures", [])
        if not isinstance(self.textures, list):
            self.textures = [self.textures]
        
        self.textures = convert_to_notify_structure(self.textures, self)
        
    @property
    def shortname(self):
        return self.id

    @shortname.setter
    def shortname(self, shortname):
        self.id = shortname
    
    @property
    def data(self):
        """
        Custom data getter allows us to re-create the json structure based
        on the saved textures.
        """
        return {
            "textures": self.textures
        }

@ImplementResource(
    ParticleFile,
    AttachableFileRP,
    EntityFileRP,
    FogFile,
    ItemFileRP,
    AnimationControllerFileRP,
    AnimationFileRP,
    MaterialFile,
    ModelFile,
    RenderControllerFile
)
@ImplementSingleResource(
    SoundsFile,
    SoundDefinitionsFile,
    FlipbookTexturesFile,
    BlocksFile,
    BiomesClientFile,
    TerrainTextureFile,
    ItemTextureFile
)
class ResourcePack(Pack):
    def __init__(self, input_path: str, project: Project = None):
        super().__init__(input_path, project=project)
        self._sounds: list[str] = []
        self._textures: list[str] = []

    @cached_property
    def sounds(self) -> list[str]:
        """
        Returns a list of all sounds in the pack, relative to the pack root.
        """
        glob_pattern = os.path.join(self.input_path, "sounds") + "/**/*."
        for extension in SOUND_EXTENSIONS:
            self._sounds.extend(glob.glob(glob_pattern + extension, recursive=True))

        self._sounds = [os.path.relpath(path, self.input_path).replace(os.sep, '/') for path in self._sounds]
        return self._sounds

    def get_sounds(self, search_path: str = "", trim_extension: bool = True) -> list[str]:
        """
        Returns a list of all child sounds of the searchpath, relative to the pack root. 
        Search path should not include 'sounds'.

        You may optionally trim the extension from the returned paths.

        Example: rp.get_sounds("entities", trim_extension=True)
        """
        sounds = []
        glob_pattern = os.path.join(self.input_path, "sounds", search_path) + "/**/*."
        for extension in SOUND_EXTENSIONS:
            sounds.extend(glob.glob(glob_pattern + extension, recursive=True))

        sounds = [os.path.relpath(path, self.input_path).replace(os.sep, '/') for path in sounds]
        if trim_extension:
            sounds = [os.path.splitext(path)[0] for path in sounds]
        return sounds
    
    @cached_property
    def textures(self) -> list[str]:
        """
        Returns a list of all textures in the pack, relative to the pack root.

        Example: "textures/my_texture.png"
        """
        glob_pattern = os.path.join(self.input_path, "textures") + "/**/*."
        for extension in TEXTURE_EXTENSIONS:
            self._textures.extend(glob.glob(glob_pattern + extension, recursive=True)) 

        self._textures = [os.path.relpath(path, self.input_path).replace(os.sep, '/') for path in self._textures]
        return self._textures
    
    def get_textures(self, search_path: str = "", trim_extension: bool = True) -> list[str]:
        """
        Returns a list of all child textures of the searchpath, relative to the pack root. 
        Search path should not include 'textures'.

        You may optionally trim the extension from the returned paths.

        Example: rp.get_textures("entities", trim_extension=True)
        """
        textures = []
        glob_pattern = os.path.join(self.input_path, "textures", search_path) + "/**/*."
        for extension in TEXTURE_EXTENSIONS:
            textures.extend(glob.glob(glob_pattern + extension, recursive=True))

        textures = [os.path.relpath(path, self.input_path).replace(os.sep, '/') for path in textures]
        if trim_extension:
            textures = [os.path.splitext(path)[0] for path in textures]
        return textures


from __future__ import annotations


"""
Module which handles core Reticulator functions. This class is generic,
and could theoretically be split into it's own library.
"""

import os
import json
import glob
import functools 

from pathlib import Path
from functools import cached_property
from typing import Union, TypeVar, Any

import dpath

NO_ARGUMENT = object()
TEXTURE_EXTENSIONS = ["png", "jpg", "jpeg", "tga"]
SOUND_EXTENSIONS = ["wav", "fsb", "ogg"]

T = TypeVar('T')

class TypeInfo:
    """
    Class for encapsulating the arguments that are passed to sub-resource based
    decorators.

    :jsonpath: The jsonpath, from where this resource can be located
    :child_cls: The class of the resource which can be accessed from this resource
        For example `AnimationRP` can be accessed from `AnimationFileRP`
    """
    
    def __init__(self, *, jsonpath="", filepath="", attribute="", getter_attribute="", extension=".json", plural=None, child_cls=None):
        self.jsonpath = jsonpath
        self.filepath = filepath
        self.attribute = attribute

        if plural == None:
            self.plural = attribute + "s"
        else:
            self.plural = plural
        self.child_cls = child_cls
        self.getter_attribute = getter_attribute
        self.extension = extension

    def __repr__(self) -> str:
        return "TypeInfo: " +  str(vars(self))

def convert_to_notify_structure(data: Union[dict, list], parent: Resource) -> Union[NotifyDict, NotifyList]:
    """
    Converts a dict or list to a notify structure.
    """
    if isinstance(data, dict):
        return NotifyDict(data, owner=parent)

    if isinstance(data, list):
        return NotifyList(data, owner=parent)

    return data




def ImplementSubResource(*args : JsonSubResource):
    """
    Wrapper class which injects all needed decorators to handle sub-resources.
    """

    def inner(parent_cls: T):
        for sub_cls in args:
            cls_type_info : TypeInfo = sub_cls.type_info

            attribute = cls_type_info.attribute
            plural = cls_type_info.plural

            @SubResourceDefinition(sub_cls)
            def x(parent_cls) -> list[T]: pass
            setattr(parent_cls, plural, x)
            x.__set_name__(parent_cls, plural)

            @DResourceGetter(sub_cls)
            def get_x(parent_cls, id: str) -> T: pass
            setattr(parent_cls, f"get_{attribute}", get_x)

            @SubResourceAdder(sub_cls)
            def add_x(parent_cls, name: str, data: dict) -> None: pass
            setattr(parent_cls, f"add_{attribute}", add_x)

        return parent_cls

    return inner

def ClassProperty(attribute: str, jsonpath: str = NO_ARGUMENT):
    """
    Class Decorator which injects a single 'property' with proper semantics.

    :param attribute: The attribute where this property can be accessed from
    :param jsonpath: The jsonpath where this attribute can be found in the json data
    """

    # Allow single-argument class-properties
    if jsonpath == NO_ARGUMENT:
        jsonpath = attribute

    def inner(cls):
        @property
        def template_property(self) -> str:
            return self.get_jsonpath(jsonpath)
        
        @template_property.setter
        def template_property(self, new_prop):
            return self.set_jsonpath(jsonpath, new_prop)

        setattr(cls, attribute, template_property)

        return cls
    return inner


def JsonChildResource(parent_cls: Resource, child_cls: T):
    """
    Special case, for looping over all children FILES to access
    RESOURCES. 
    
    Classic example is doing `rp.animations` instead of looping over all
    animation files and then painfully getting animations from there.

    NOT CACHED
    """
    
    parent_attribute = parent_cls.type_info.plural
    child_attribute = child_cls.type_info.plural

    def decorator(func) -> cached_property[list[T]]:
        @property
        @functools.wraps(func)
        def wrapper(self) -> list[T]:
            sub_resources = []
            for file_resource in getattr(self, parent_attribute):
                for sub_resource in getattr(file_resource, child_attribute):
                    sub_resources.append(sub_resource)
            return sub_resources
        return wrapper
    return decorator


def ImplementSingleResource(*args : JsonFileResource):
    """
    Wrapper decorator for implementing multiple Single Resources.
    """

    def inner(parent_cls):
        for sub_cls in args:
            cls_type_info : TypeInfo = sub_cls.type_info

            attribute = cls_type_info.attribute

            @DSingleResourceDefinition(sub_cls)
            def x(parent_cls): pass
            setattr(parent_cls, attribute, x)
            x.__set_name__(parent_cls, attribute)

        return parent_cls

    return inner

def DSingleResourceDefinition(cls : T):
    """Inserts implementation for JsonFileResource (single)"""

    attribute = cls.__name__
    filepath = cls.type_info.filepath
    extension = cls.type_info.extension

    def decorator(func) -> cached_property[T]:
        @cached_property
        @functools.wraps(func)
        def wrapper(self) -> T:
            new_object = cls(filepath = filepath + extension, pack = self)
            setattr(self, attribute, new_object)
            return new_object
        return wrapper
    return decorator


def SubResourceDefinition(cls: T):
    jsonpath = cls.type_info.jsonpath
    attribute_plural = cls.type_info.plural

    def decorator(func) -> cached_property[list[T]]:
        @cached_property
        @functools.wraps(func)
        def wrapper(self) -> list[T]:
            setattr(self, attribute_plural, [])
            for path, data in self.get_data_at(jsonpath):
                getattr(self, attribute_plural).append(cls(parent = self, json_path = path, data = data))
            return getattr(self, attribute_plural)
        return wrapper
    return decorator



#region Resources


def ImplementResource(*args : JsonFileResource):
    """
    Class Decorator which interjects functions to deal with resources. Wraps:
      - DResourceDefinition
      - DResourceGetter
      - DResourceGetter
    """
        
    def inner(parent_cls):
        for sub_cls in args:
            cls_type_info : TypeInfo = sub_cls.type_info

            attribute = cls_type_info.attribute
            plural = cls_type_info.plural

            @DResourceDefinition(sub_cls)
            def x(parent_cls): pass
            setattr(parent_cls, plural, x)
            x.__set_name__(parent_cls, plural) # See: CachedProperty docs.

            @DResourceGetter(sub_cls)
            def get_x(parent_cls, id: str): pass
            setattr(parent_cls, f"get_{attribute}", get_x)

            @DResourceAdder(sub_cls)
            def add_x(parent_cls, filepath: str, data: dict): pass
            setattr(parent_cls, f"add_{attribute}", add_x)

            child_cls = cls_type_info.child_cls
            if child_cls != None:
                @JsonChildResource(sub_cls, child_cls)
                def child_x(self): pass
                setattr(parent_cls, child_cls.type_info.plural, child_x)

                @ChildGetter(sub_cls, child_cls)
                def get_child_x(self, id: str): pass
                setattr(parent_cls, f"get_{child_cls.type_info.attribute}", get_child_x)

        return parent_cls

    return inner

def DResourceDefinition(cls : T):
    """
    Decorator which implements support for a JsonFileResource.
    """

    attribute = cls.__name__
    filepath = cls.type_info.filepath
    extension = cls.type_info.extension

    def decorator(func) -> cached_property[list[T]]:
        @cached_property
        @functools.wraps(func)
        def wrapper(self) -> list[T]:
            setattr(self, attribute, [])
            base_directory = os.path.join(self.input_path, filepath)

            for local_path in glob.glob(base_directory + "/**/*" + extension, recursive=True):
                local_path = os.path.relpath(local_path, self.input_path)

                getattr(self, attribute).append(cls(filepath = local_path, pack = self))
            return getattr(self, attribute)
        return wrapper
    return decorator

def DResourceGetter(cls : T):
    """
    Decorator which allows you to get a resource. For example getting a component
    from an entity.
    """
    attribute_plural = cls.type_info.plural
    getter_attribute = cls.type_info.getter_attribute
    def decorator(func) -> T:
        @functools.wraps(func)
        def wrapper(self, compare):
            for child in getattr(self, attribute_plural):
                if smart_compare(getattr(child, getter_attribute), compare):
                    return child
            return None # No longer raises an error. Allow a getter to return none.
        return wrapper
    return decorator

def DResourceAdder(cls : Resource):
    """
    This decorator allows you to inject SubResources into your Resources.
    """
    base_filepath = cls.type_info.filepath
    attribute = cls.type_info.plural

    def decorator(func):
        @functools.wraps(func)
        def wrapper(self, filepath, data):
            new_filepath = base_filepath + "/" + filepath
            new_object = cls(data=data, pack=self, filepath=new_filepath)

            if new_object:
                getattr(self, attribute).append(new_object)
                return new_object
            else:
                raise ReticulatorException()

        return wrapper
    return decorator

#endregion

def ChildGetter(parent_cls : Any, child_cls : T):
    """
    Special getter for getting sub-resources that are further down the chain.
    For example getting 'animations' directly from the RP without passing
    through the AnimationFile class.
    """
    parent_attribute = parent_cls.type_info.plural
    child_attribute = child_cls.type_info.plural
    getter_attribute = child_cls.type_info.getter_attribute
    def decorator(func) -> T:
        @functools.wraps(func)
        def wrapper(self, compare):
            for child in getattr(self, parent_attribute):
                for grandchild in getattr(child, child_attribute):
                    if smart_compare(getattr(grandchild, getter_attribute), compare):
                        return grandchild
                return None
        return wrapper
    return decorator

def SubResourceAdder(cls : Resource):
    """
    This decorator allows you to inject SubResources into your Resources.
    """
    jsonpath = cls.type_info.jsonpath
    attribute = cls.type_info.plural

    def decorator_sub_resource(func):
        @functools.wraps(func)
        def wrapper_sub_resource(self, *args, **kwargs):
            # This handles the case where a class is passed in directly.
            # like `add_box(Box(..))``

            if len(args) > 0:
                raise ReticulatorException("This function can only be called with keyword arguments: 'resource' (alone), or 'id' and 'data'")

            if kwargs.get('resource') == None and (kwargs.get('id') == None or kwargs.get('data') == None):
                raise ReticulatorException("This function can only be called with 'resource' OR 'id and 'data'.")

            # Handle Object case
            if new_object := kwargs.get('resource'):
                self.set_jsonpath(new_object.json_path, new_object.data)
                new_object.parent = self

            # This handles the 'normal' flow, where arguments are passed in
            # via 'parts' and are constructed automatically.
            else:
                id = kwargs.get('id')
                data = kwargs.get('data')

                if id == None:
                    raise ReticulatorException("Id may not be None.")

                if data == None:
                    raise ReticulatorException("Data may not be None")

                new_jsonpath = jsonpath + "/" + id
                self.set_jsonpath(new_jsonpath, data)
                new_object = cls(data=data, parent=self, json_path=new_jsonpath)

            if new_object:
                getattr(self, attribute).append(new_object)
                return new_object
            else:
                raise ReticulatorException()


        return wrapper_sub_resource
    return decorator_sub_resource

# Methods
def create_nested_directory(path: str):
    """
    Creates a nested directory structure if it doesn't exist.
    """
    Path(path).parent.mkdir(parents=True, exist_ok=True)

def save_json(filepath, data):
    """
    Saves json to a filepath, creating nested directory if required.
    """
    create_nested_directory(filepath)
    with open(filepath, "w+") as file_head:
        return json.dump(data, file_head, indent=2, ensure_ascii=False)

def smart_compare(a, b) -> bool:
    """
    Compares to objects using ==, but if they can both be interpreted as
    path-like objects, it uses path comparison.
    """

    try:
        a = Path(a)
        b = Path(b)

        return a == b
    except Exception:
        return a == b


# Exceptions
class ReticulatorException(Exception):
    """
    Base class for Reticulator exceptions.
    """

class FloatingAssetError(ReticulatorException):
    """
    Called when a "floating" asset attempts to access its parent pack, for
    example when saving
    """

class AssetNotFoundError(ReticulatorException):
    """
    Called when attempting to access an asset that does not exist, for
    example getting an entity by name
    """

class InvalidJsonError(ReticulatorException):
    """
    Called when a JSON file is invalid.
    """

class AmbiguousAssetError(ReticulatorException):
    """
    Called when a path is not unique.
    """

class FormatVersionError(ReticulatorException):
    """
    Called when a format version does not exist for a file
    """


# Base Classes

class NotifyDict(dict):
    """
    A notify dictionary is a dictionary that can notify its parent when its been
    edited.
    """
    def __init__(self, *args, owner: Resource = None, **kwargs):
        self._owner = owner

        if len(args) > 0:
            for key, value in args[0].items():
                args[0][key] = convert_to_notify_structure(value, self._owner)
        super().__init__(*args, **kwargs)

    def get_item(self, attr):
        try:
            return self.__getitem__(attr)
        except Exception:
            return None
    
    def __delitem__(self, v) -> None:
        if(self._owner != None):
            self._owner.dirty = True
        return super().__delitem__(v)

    def __setitem__(self, attr, value):
        value = convert_to_notify_structure(value, self._owner)

        if self.get_item(attr) != value:
            if(self._owner != None):
                self._owner.dirty = True

        super().__setitem__(attr, value)

class NotifyList(list):
    """
    A notify list is a list which can notify its owner when it has
    changed.
    """
    def __init__(self, *args, owner: Resource = None, **kwargs):
        self._owner = owner

        if len(args) > 0:
            for i in range(len(args[0])):
                args[0][i] = convert_to_notify_structure(args[0][i], self._owner)

        super().__init__(*args, **kwargs)

    def get_item(self, attr):
        try:
            return self.__getitem__(attr)
        except Exception:
            return None
    
    def __delitem__(self, v) -> None:
        if(self._owner != None):
            self._owner.dirty = True
        return super().__delitem__(v)
    
    def append(self, v):
        if(self._owner != None):
            self._owner.dirty = True
        super().append(v)

    def extend(self, v):
        if(self._owner != None):
            self._owner.dirty = True
        super().extend(v)

    def __setitem__(self, attr, value):
        value = convert_to_notify_structure(value, self._owner)

        if self.__getitem__(attr) != value:
            if(self._owner != None):
                self._owner.dirty = True
        
        super().__setitem__(attr, value)


class Resource():
    """
    The top resource in the inheritance chain.

    Contains:
     - reference to the Pack (could be blank, if floating resource)
     - reference to the File (could be a few links up the chain)
     - dirty status
     - abstract ability to save, including context manager support
     - list of children resources
    """
    
    jsonpath : str
    
    def __init__(self, file: FileResource = None, pack: Pack = None) -> None:
        # Public
        self.pack: Union[ResourcePack, BehaviorPack] = pack
        self.file = file
        self._dirty = False
        self._deleted = False

        # Private
        self._resources: Resource = []

    def __enter__(self) -> Resource:
        """
        Context manager support.
        """
        return self

    def __exit__(self, type, value, traceback):
        """
        Context manager support. Will save the resource on exit.
        """
        self.save()
    
    @property
    def dirty(self):
        """
        Whether the asset is dirty (has been edited).
        """
        return self._dirty

    @dirty.setter
    def dirty(self, dirty):
        self._dirty = dirty

    def register_resource(self, resource) -> None:
        """
        Register a child resource. These resources will always be saved first.
        """
        self._resources.append(resource)

    def _save(self):
        """
        Internal implementation of asset saving.

        Should always be implemented.
        """
        raise NotImplementedError()

    def _delete(self):
        """
        Internal implementation of asset deletion.

        Should always be implemented.
        """
        raise NotImplementedError()


    def save(self, force=False):
        """
        Saves the resource. Should not be overridden.

        This method will save all child resources first, by calling
        the internal _save implementation.
        """

        # File Assets without packs may not save.
        if isinstance(self, FileResource) and not self.pack:
            raise FloatingAssetError("Assets without a pack cannot be saved.")

        # Only dirty assets can be saved, unless forced.
        if self.dirty or force:
            # Save all children first
            for resource in self._resources:
                resource.save(force=force)

            # Now, save this resource
            self._save()

            # Mark as clean
            self.dirty = False


    def delete(self):
        """
        Deletes the resource. Should not be overridden.
        """

        # First, delete all resources of children
        for resource in self._resources:
            resource.delete()

        # Then delete self
        self._delete()


class FileResource(Resource):
    """
    A resource, which is also a file.
    Contains:
     - File path
     - Ability to mark for deletion
    """
    def __init__(self, filepath: str = None, pack: Pack = None) -> None:
        # Initialize resource
        super().__init__(file=self, pack=pack)

        # All files must register themselves with their pack
        # if the pack exists.
        if self.pack:
            pack.register_resource(self)

        # Public
        self.filepath = filepath

        if filepath:
            self.file_name = os.path.basename(filepath)

    def _delete(self):
        """
        Internal implementation of file deletion.

        Marking for deletion allows us to delete the file during saving,
        without effecting the file system during normal operation.
        """
        self._deleted = True

class JsonResource(Resource):
    """
    Parent class, which is responsible for all resources which contain
    json data.
    Should not be used directly. Use JsonFileResource, or JsonSubResource instead.
    Contains:
     - Data object
     - Method for interacting with the data
    """

    # The type information, used for generating this class at runtime.
    type_info : TypeInfo
    
    def __init__(self, data: dict = None, file: FileResource = None, pack: Pack = None) -> None:
        super().__init__(file=file, pack=pack)
        self._data = data
    
    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, data):
        self.dirty = True
        self._data = data

    def _save(self):
        raise NotImplementedError("This json resource cannot be saved.")

    def _delete(self):
        raise NotImplementedError("This json resource cannot be deleted.")

    def __str__(self):
        return json.dumps(self.data, indent=2, ensure_ascii=False)

    def jsonpath_exists(self, json_path:str) -> bool:
        """
        Checks if a jsonpath exists
        """
        try:
            self.get_jsonpath(json_path)
            return True
        except AssetNotFoundError:
            return False

    def delete_jsonpath(self, json_path:str) -> None:
        """
        Removes value at jsonpath location.
        """
        path_exists = self.jsonpath_exists(json_path)
        if path_exists:
            self.dirty = True
            dpath.delete(self.data, json_path)

    def pop_jsonpath(self, json_path, default=NO_ARGUMENT) \
        -> Union[dict, list, int, str, float]:
        """
        Removes value at jsonpath location, and returns it.
        """

        data = self.get_jsonpath(json_path, default=default)
        self.delete_jsonpath(json_path)
        self.dirty = True
        return data

    def append_jsonpath(self, json_path:str, insert_value:Any):
        """
        Appends a value at jsonpath location. Will create path if it doesn't exist.
        """

        path_exists = self.jsonpath_exists(json_path)

        # Otherwise, set the value
        self.dirty = True

        if path_exists:
            self.get_jsonpath(json_path).append(insert_value)
        else:
            self.set_jsonpath(json_path, [insert_value])

    def set_jsonpath(self, json_path:str, insert_value:any, overwrite:bool=True):
        """
        Sets value at jsonpath location.

        Can create a new key if it doesn't exist.
        """

        path_exists = self.jsonpath_exists(json_path)

        # If overwrite is false, it will set the path only
        # if there is no path.
        if path_exists and not overwrite:
            return

        # Otherwise, set the value
        self.dirty = True
        dpath.new(self.data, json_path, insert_value)
        

    def get_jsonpath(self, json_path, default=NO_ARGUMENT):
        """
        Gets value at jsonpath location.

        A default value may be provided, for missing keys.

        raises:
            AssetNotFoundError if the path does not exist.
        """
        try:
            return dpath.get(self.data, json_path)
        except Exception as exception:
            if default is not NO_ARGUMENT:
                return default
            raise AssetNotFoundError(
                f"Path {json_path} does not exist."
            ) from exception

    def get_data_at(self, json_path):
        """
        Returns a list of jsonpaths found at this jsonpath location.
        For a list, this will return: json_path + list index
        For a dict, this will return json_path + dict key

        Exception will be raised if the result is not a list or dict.
        Missing data path will return []
        """
        try:
            # Special case, to handle the root of the json
            if json_path == "**":
                result = self.data
            else:
                result = self.get_jsonpath(json_path, default=[])

            if isinstance(result, dict):
                for key in result.keys():
                    yield json_path + f"/{key}", result[key]
            elif isinstance(result, list):
                for i, element in enumerate(result):
                    yield json_path + f"/[{i}]", element
            else:
                raise AmbiguousAssetError(f"Path '{json_path}' matched a single element, not a list or dict.")

        except KeyError as key_error:
            raise AssetNotFoundError(json_path, self.data) from key_error


class JsonSubResource(JsonResource):
    """
    A sub resource represents a chunk of json data, within a file.
    """
    def __init__(self, parent: Resource = None, json_path: str = None, data: dict = None) -> None:
        super().__init__(data = data, pack = parent.pack, file = parent.file)

        # The parent of the sub-resource is the asset of type Resource
        # which owns this sub-resource. For example a Component is owned by
        # either an EntityFileBP, or a ComponentGroup.
        self.parent: JsonResource = parent

        # The jsonpath is the location within the parent resource, where
        # this sub-resource is stored.
        self._json_path: str = json_path

        # The original path location needs to be stored, for the purpose
        # of renaming or moving the sub-resource.
        self._original_json_path: str = json_path

        # Register self into parent, so that it can be found by the parent
        # during saving, etc.
        self.parent.register_resource(self)
    
    @property
    def json_path(self):
        return self._json_path

    @json_path.setter
    def json_path(self, json_path):
        self.dirty = True
        self._json_path = json_path

    @property
    def id(self):
        """
        The ID of the sub-resource, such as 'minecraft:scale' for a component.
        """
        return self.json_path.rsplit("/", maxsplit=1)[1]

    @id.setter
    def id(self, id):
        self.dirty = True
        self.json_path = self.json_path.rsplit("/", maxsplit=1)[0] + "/" + id

    def __repr__(self):
        return f"'{self.__class__.__name__}: {self.id}'"

    def __str__(self):
        """
        Returns string representation of this resource.

        The json data will be representation for printing: With the id appended,
        and with indentation.
        """
        return f'"{self.id}": {json.dumps(self.data, indent=2, ensure_ascii=False)}'

    @property
    def dirty(self):
        """
        Returns True if this resource has been modified since it was last saved.
        """
        return self._dirty

    @dirty.setter
    def dirty(self, dirty):
        """
        Set the dirty flag, which stores whether this resource has been
        modified since it was last saved.
        """
        self._dirty = dirty
        self.parent.dirty = dirty # Pseudo recursive call will travel up the chain


    def _save(self):
        """
        Saves the resource, into its parents json structure.

        This works by replacing the data at the jsonpath location,
        meaning that the parent will contain accurate representation of
        the children's data, saved into itself.
        """

        # If the id was updated, we need to serialize into a new location.
        # This implies first deleting the old location, and then creating
        # a new one.
        if not self._deleted:
            if self.json_path != self._original_json_path:
                self.parent.delete_jsonpath(self._original_json_path)
                self.parent.set_jsonpath(self.json_path, self.data)
            else:
                self.parent.set_jsonpath(self.json_path, self.data)

        self._dirty = False

    def _delete(self):
        """
        Deletes itself from parent, by removing the data at the jsonpath
        location.

        If it is a list, it must only set itself to None. The actual
        deletion will take place during saving.
        """
        self.parent.dirty = True
        self.parent._resources.remove(self)
        self._deleted = True
        self.parent.delete_jsonpath(self.json_path)

class JsonFileResource(FileResource, JsonResource):
    """
    A file, which contains json data. Most files in the addon system
    are of this type, or have it as a resource parent.
    """
    def __init__(self, data: dict = None, filepath: str = None, pack: Pack = None) -> None:
        # Init file resource parent, which gives us access to data
        FileResource.__init__(self, filepath=filepath, pack=pack)
        
        # Data is either set directly, or is read from the filepath for this
        # resource. This allows assets to be created from scratch, whilst
        # still having an associated file location.
        if data is not None:
            self.data = data
        else:
            self.data = self.load_json(self.filepath)

        # Init json resource, which relies on the new data attribute
        JsonResource.__init__(self, data=self.data, file=self, pack=pack)

    def __repr__(self):
        return f"'{self.__class__.__name__}: {self.filepath}'"


    def load_json(self, filepath: str) -> dict:
        """
        Loads json from file. `local_path` paramater is local to the projects
        input path.
        """
        
        if self.pack:
            filepath = os.path.join(self.pack.input_path, filepath)

        if not os.path.exists(filepath):
            raise AssetNotFoundError(f"File not found: {filepath}")
        try:
            with open(filepath, "r", encoding='utf8') as fh:
                try:
                    return json.load(fh)
                except json.JSONDecodeError:
                    try:
                        fh.seek(0)
                        contents = ""
                        for line in fh.readlines():
                            cleaned_line = line.split("//", 1)[0]
                            if len(cleaned_line) > 0 and line.endswith("\n") and "\n" not in cleaned_line:
                                cleaned_line += "\n"
                            contents += cleaned_line
                        while "/*" in contents:
                            pre_comment, post_comment = contents.split("/*", 1)
                            contents = pre_comment + post_comment.split("*/", 1)[1]
                        return json.loads(contents)
                    except json.JSONDecodeError as exception:
                        return {}
        except Exception:
            raise InvalidJsonError(filepath)

    def _save(self):
        save_path = os.path.join(self.pack.output_directory, self.filepath)
        create_nested_directory(save_path)

        # If the file has been marked for deletion, delete it
        if self._deleted:
            # If the paths are the same, delete the file, otherwise
            # we can just pass
            if smart_compare(self.pack.input_path, self.pack.output_directory):
                os.remove(save_path)
        else:
            with open(save_path, "w+") as file_head:
                clean_data = {k: v for k, v in self.data.items() if v is not None}
                return json.dump(clean_data, file_head, indent=2, ensure_ascii=False)